# Copyright 2009-2011 Yelp
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Write and run Hadoop Streaming jobs on Amazon Elastic MapReduce or your own Hadoop cluster.
"""

__author__ = 'David Marin <dave@yelp.com>'

__credits__ = ['James Brown <jbrown@yelp.com>',
               'Adam Derewicki <derwiki@yelp.com>',
               'Peter Harrington <peter.b.harrington@gmail.com>',
               'Brett Hoerner <brett@bretthoerner.com>',
               'Matt Jones <mattj@yelp.com>',
               'Julian Krause <juliank@yelp.com>',
               'Jimmy Retzlaff <jretz@yelp.com>',
               'Ned Rockson <ned@tellapart.com>',
               'Paul Wais <pwais@yelp.com>']

__version__ = '0.2.4'
